
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#include <climits>		// ULONG_MAX
#include <cmath>		// log
#include <cstdlib>		// srand, rand, RAND_MAX
#include <ctime>		// time

#include "Random.hh"


namespace Mf {


unsigned setSeed(unsigned theSeed)
{
	srand(theSeed);
	return theSeed;
}

unsigned setSeed()
{
	return setSeed(time(0));
}


template <typename T>
T getRandom()
{
	const  float ln2 = 0.693147;
	static const unsigned bitsPerPiece = std::log(float(RAND_MAX)) / ln2;
	static const unsigned sizeOfType = sizeof(T) * 8;
	static const unsigned pieces = sizeOfType / bitsPerPiece +
								 ((sizeOfType % bitsPerPiece) ? 1 : 0);

	T bits = 0;

	// we need to call rand() multiple times if it won't provide enough random
	// bits to fill the size of the given type
	for (unsigned i = 0; i < pieces; i++)
	{
		long piece = rand();
		bits |= piece << (i * bitsPerPiece);
	}

	return bits;
}


template <>
bool getRandom<bool>()
{
	return rand() % 2;
}

template <typename T>
T getRandom(T lower, T upper)
{
	unsigned short randInt = getRandom<unsigned int>();
	return lower + T(randInt % (upper - lower + 1));
}

template <>
float getRandom(float lower, float upper)
{
	unsigned long randInt = getRandom<unsigned long>();
	return (float(randInt) / float(ULONG_MAX)) * (upper - lower) + lower;
}

template <>
double getRandom(double lower, double upper)
{
	unsigned long long randInt = getRandom<unsigned long long>();
	return (double(randInt) / double(ULLONG_MAX)) * (upper - lower) + lower;
}

template <>
float getRandom()
{
	return getRandom<float>(0.0, 1.0);
}

template <>
double getRandom()
{
	return getRandom<double>(0.0, 1.0);
}


}; // namespace Mf

/** vim: set ts=4 sw=4 tw=80: *************************************************/

