
/*******************************************************************************

 Copyright (c) 2009, Charles McGarvey
 All rights reserved.
 
 Redistribution   and   use  in  source  and  binary  forms,  with  or  without
 modification, are permitted provided that the following conditions are met:
 
   * Redistributions  of  source  code  must retain the above copyright notice,
     this list of conditions and the following disclaimer.
   * Redistributions  in binary form must reproduce the above copyright notice,
     this  list of conditions and the following disclaimer in the documentation
     and/or other materials provided with the distribution.
 
 THIS  SOFTWARE  IS  PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND  ANY  EXPRESS  OR  IMPLIED  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 DISCLAIMED.  IN  NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 FOR  ANY  DIRECT,  INDIRECT,  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 DAMAGES  (INCLUDING,  BUT  NOT  LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 SERVICES;  LOSS  OF  USE,  DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 CAUSED  AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 OR  TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*******************************************************************************/

#ifndef _MOOF_SINGLETON_HH_
#define _MOOF_SINGLETON_HH_

#include <stdexcept>


namespace Mf {


template <typename T>
class Singleton
{
	static T* ptr_;

public:
	struct Exception : public std::runtime_error
	{
		explicit Exception(const std::string& what_arg) :
			std::runtime_error(what_arg) {}
	};

	Singleton()
	{
		if (!ptr_)
		{
			// This hack is from Game Programming Gems.
			long long offset = (long long)(T*)1 - (long long)(Singleton<T>*)(T*)1;
			ptr_ = (T*)((long long)this + offset);
		}
	}
	~Singleton()
	{
		long long offset = (long long)(T*)1 - (long long)(Singleton<T>*)(T*)1;
		if (ptr_ == (T*)((long long)this + offset))
		{
			ptr_ = 0;
		}
	}

	static T& instance()
	{
		if (!ptr_)
		{
			throw Exception("accessing uninstantiated singleton");
		}
		return *ptr_;
	}

	static T* instance_ptr()
	{
		return ptr_;
	}
};

template <typename T> T* Singleton<T>::ptr_ = 0;


} // namespace Mf

#endif // _MOOF_SINGLETON_HH_

/** vim: set ts=4 sw=4 tw=80: *************************************************/

